package java.util.concurrent;

import sun.misc.Unsafe;

public abstract class CountedCompleter<T> extends ForkJoinTask<T>
{
    private static final long serialVersionUID = 5232453752276485070L;
    final CountedCompleter<?> completer;
    volatile int pending;
    private static final Unsafe U;
    private static final long PENDING;
    
    protected CountedCompleter(CountedCompleter<?> completer, int pending) {
        this.completer = completer;
        this.pending = pending;
    }
    
    protected CountedCompleter(CountedCompleter<?> completer) {
        this.completer = completer;
    }
    
    protected CountedCompleter() {
        this.completer = null;
    }
    
    public abstract void compute();
    
    public void onCompletion(CountedCompleter<?> countedCompleter) {
    }
    
    public boolean onExceptionalCompletion(Throwable t, CountedCompleter<?> countedCompleter) {
        return true;
    }
    
    public final CountedCompleter<?> getCompleter() {
        return this.completer;
    }
    
    public final int getPendingCount() {
        return this.pending;
    }
    
    public final void setPendingCount(int pending) {
        this.pending = pending;
    }
    
    public final void addToPendingCount(int delta) {
        CountedCompleter.U.getAndAddInt(this, CountedCompleter.PENDING, delta);
    }
    
    public final boolean compareAndSetPendingCount(int expected, int x) {
        return CountedCompleter.U.compareAndSwapInt(this, CountedCompleter.PENDING, expected, x);
    }
    
    public final int decrementPendingCountUnlessZero() {
        int pending;
        while ((pending = this.pending) != 0 && !CountedCompleter.U.compareAndSwapInt(this, CountedCompleter.PENDING, pending, pending - 1)) {}
        return pending;
    }
    
    public final CountedCompleter<?> getRoot() {
        CountedCompleter<Object> completer;
        while ((completer = this.completer) != null) {
            this = completer;
        }
        return this;
    }
    
    public final void tryComplete() {
        this = (this = this);
        while (true) {
            int pending;
            if ((pending = this.pending) == 0) {
                this.onCompletion(this);
                if ((this = (this = this).completer) != null) {
                    continue;
                }
                this.quietlyComplete();
                return;
            }
            else {
                if (!CountedCompleter.U.compareAndSwapInt(this, CountedCompleter.PENDING, pending, pending - 1)) {
                    continue;
                }
                return;
            }
        }
    }
    
    public final void propagateCompletion() {
        this = (this = this);
        while (true) {
            int pending;
            if ((pending = this.pending) == 0) {
                CountedCompleter countedCompleter;
                if ((this = (countedCompleter = this).completer) != null) {
                    continue;
                }
                countedCompleter.quietlyComplete();
                return;
            }
            else {
                if (!CountedCompleter.U.compareAndSwapInt(this, CountedCompleter.PENDING, pending, pending - 1)) {
                    continue;
                }
                return;
            }
        }
    }
    
    @Override
    public void complete(T rawResult) {
        this.setRawResult(rawResult);
        this.onCompletion(this);
        this.quietlyComplete();
        CountedCompleter<Object> completer;
        if ((completer = this.completer) != null) {
            completer.tryComplete();
        }
    }
    
    public final CountedCompleter<?> firstComplete() {
        int pending;
        while ((pending = this.pending) != 0) {
            if (!CountedCompleter.U.compareAndSwapInt(this, CountedCompleter.PENDING, pending, pending - 1)) {
                continue;
            }
            return null;
        }
        return this;
    }
    
    public final CountedCompleter<?> nextComplete() {
        CountedCompleter<Object> completer;
        if ((completer = this.completer) != null) {
            return completer.firstComplete();
        }
        this.quietlyComplete();
        return null;
    }
    
    public final void quietlyCompleteRoot() {
        CountedCompleter<Object> completer;
        while ((completer = this.completer) != null) {
            this = completer;
        }
        this.quietlyComplete();
    }
    
    public final void helpComplete(int n) {
        if (n > 0 && this.status >= 0) {
            Thread currentThread;
            if (((currentThread = Thread.currentThread()) instanceof ForkJoinWorkerThread)) {
                ForkJoinWorkerThread forkJoinWorkerThread = (ForkJoinWorkerThread)currentThread;
                forkJoinWorkerThread.pool.helpComplete(forkJoinWorkerThread.workQueue, this, n);
            }
            else {
                ForkJoinPool.common.externalHelpComplete(this, n);
            }
        }
    }
    
    @Override
    void internalPropagateException(Throwable t) {
        this = (this = this);
        while (this.onExceptionalCompletion(t, this) && (this = (this = this).completer) != null && this.status >= 0 && this.recordExceptionalCompletion(t) == Integer.MIN_VALUE) {}
    }
    
    @Override
    protected final boolean exec() {
        this.compute();
        return false;
    }
    
    @Override
    public T getRawResult() {
        return null;
    }
    
    @Override
    protected void setRawResult(T t) {
    }
    
    static {
        try {
            U = Unsafe.getUnsafe();
            PENDING = CountedCompleter.U.objectFieldOffset(CountedCompleter.class.getDeclaredField("pending"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
